#!/usr/bin/perl
use strict;

################### Functions used in the script ########################
sub sum{
	my $sum=0;
	foreach(@_){
		$sum+=$_;
	}
	return $sum;
}

sub max{
	my $max=undef;
	foreach(@_){
		if($_ eq "Inf"){
			$max="Inf";
			last;
		}
		$max=$_ if((! defined $max) || ($_>$max));
	}
	return $max;
}

sub min{
	my $min=undef;
	foreach(@_){
		next if($_ eq "Inf");
		$min=$_ if((! defined $min) || ($_<$min));
	}
	$min="Inf" if(@_>1 and (! defined $min));
	return $min;
}

sub which_max{
	my $max=undef;
	my $index_max=-1;
	for(my $i=0; $i<@_; $i++){
		if($_ eq "Inf"){
			$index_max=$i;
			last;
		}
		if((! defined $max) || ($_[$i]>$max)){
			$index_max=$i;
			$max=$_[$i];
		}
	}
	return $index_max;
}

sub which_min{
	my $min=undef;
	my $index_min=-1;
	for(my $i=0; $i<@_; $i++){
		if($_[$i] eq "Inf"){
			next;
		}
		if((! defined $min) || ($_[$i]<$min)){
			$index_min=$i;
			$min=$_[$i];
		}
	}
	return $index_min;
}

sub mean{
	my $mean=0;
	for(my $i=0; $i<@_; $i++){
		if($_[$i] eq "Inf"){
			return "Inf";
		}
		$mean+=$_[$i]/(scalar @_);
	}
	return $mean;
}

sub sd{
	my $sd=0;
	my $mean=mean(@_);
	if($mean eq "Inf"){
		return "NA";
	}

	for(my $i=0; $i<@_; $i++){
		$sd+=($_[$i]-$mean)**2;
	}
	$sd/=@_-1;
	$sd=$sd**0.5;

	return $sd;
}

sub time_warping{
	my @series1=@{$_[0]};
	my @series2=@{$_[1]};
	my $w_size=$_[2];
	my $idx=(defined $_[3]) ? $_[3] : -1;
	my $num_markers=($idx==-1) ? scalar @{$series1[0]} : 1;
	
	# determine weight of each marker profile
	my @weights=();
	if($idx==-1 and defined $_[4]){
		@weights=@{$_[4]};
	}
	elsif($idx==-1){
		for(my $i=0; $i<$num_markers; $i++){
			push(@weights, 1);
		}
	}

	# determine peak (high or low) of each marker profile at each condition
	my @peaks=();
	if($idx==-1){
		for(my $i=0; $i<$num_markers; $i++){
			my %peaks_marker_c1=();
			my %peaks_marker_c2=();

			my @profile1=();
			my @profile2=();
			for(my $j=0; $j<@series1; $j++){
				push(@profile1, $series1[$j][$i]);
			}
			for(my $j=0; $j<@series2; $j++){
				push(@profile2, $series2[$j][$i]);
			}

			my $mean1=mean(@profile1); my $sd1=sd(@profile1);
			my $mean2=mean(@profile2); my $sd2=sd(@profile2);
			#warn "$i\t$mean1\t$mean2\t$sd1\t$sd2\n";
			for(my $j=0; $j<@profile1; $j++){
				#$peaks_marker_c1{$j}++ if($profile1[$j]>$mean1+$sd1 || $profile1[$j]<$mean1-$sd1);
				$peaks_marker_c1{$j}++ if($j==which_max(@profile1) || $j==which_min(@profile1))
			}
			for(my $j=0; $j<@profile2; $j++){
				#$peaks_marker_c2{$j}++ if($profile2[$j]>$mean2+$sd2 || $profile2[$j]<$mean2-$sd2);
				$peaks_marker_c2{$j}++ if($j==which_max(@profile2) || $j==which_min(@profile2))
			}
			
			my @res=(\%peaks_marker_c1, \%peaks_marker_c2);
			$peaks[$i]=\@res;
		}
	}
	
	# initialize distance and path matrix
	my @dist=();
	my @path=();
	for(my $i=0; $i<=@series1; $i++){
		for(my $j=0; $j<=@series2; $j++){
			$dist[$i][$j]="Inf";
			$path[$i][$j]=-1;
		}
	}
	$dist[0][0]=0;

	# estimate distance and path matrix
	for(my $i=1; $i<=@series1; $i++){
		for(my $j=max(1, $i-$w_size); $j<=min(scalar @series2, $i+$w_size); $j++){
			my @vector1=@{$series1[$i-1]};
			my @vector2=@{$series2[$j-1]};

			my $cost=0;
			if($idx==-1){
				my $here_has_a_peak=0;
				for(my $k=0; $k<$num_markers; $k++){
					if(exists $peaks[$k][0]{$i-1} && $peaks[$k][1]{$j-1}){
						$here_has_a_peak=1;
						last;
					}
				}
				if($here_has_a_peak==0){
					for(my $k=0; $k<$num_markers; $k++){
						$cost+=$weights[$k] * (($vector1[$k]-$vector2[$k])**2)**0.5;
					}
				}
			}
			else{
				$cost=(($vector1[$idx]-$vector2[$idx])**2)**0.5;
			}

			# path 0 [deletion]: j-th of series2 is correponding to both i-th and (i-1)th of series1
			# path 1 [insertion]: both j-th and (j-1)th of series2 is corresponding to i-th of series1)
			# path 2 [match]: j-th of series 2 is correponding to i-th of series 1, while (j-1)th of series2 is to (i-1) of series1
			$dist[$i][$j]=$cost+min($dist[$i-1][$j], $dist[$i][$j-1], $dist[$i-1][$j-1]); 
			$path[$i][$j]=which_min($dist[$i-1][$j], $dist[$i][$j-1], $dist[$i-1][$j-1]);
		}
	}
		
	## get the path by tracing back ##
	my ($i, $j)=(scalar @series1, scalar @series2);
	my @aligned;
	while($i>0 && $j>0){
		push(@{$aligned[0]}, $i); push(@{$aligned[1]}, $j);
		if($path[$i][$j]==0){
			$i--;
		}
		elsif($path[$i][$j]==1){
			$j--;
		}
		elsif($path[$i][$j]==2){
			$i--; $j--;
		}
	}
	@{$aligned[0]}=reverse @{$aligned[0]};
	@{$aligned[1]}=reverse @{$aligned[1]};
	
	return(\@aligned);
}

sub align{ ### the main function to do alignment
	my @series1=@{$_[0]};
	my @series2=@{$_[1]};
	my @weights=@{$_[2]};
	my $align_mode=(defined $_[3]) ? $_[3] : 1; # alignment mode. 0: one by one and then intergrate; 1: altogether when calculating cost
	my $w_size=(defined $_[4]) ? $_[4] : max(scalar @series1, scalar @series2)/2;

	# mode 0: do section alignment for each marker separately, and then intergrate them together
	if($align_mode==0){ 
		### initial path count ###
		my @path_count=();
		for(my $i=1; $i<=@series1; $i++){
			for(my $j=1; $j<=@series2; $j++){
				my @counts=(0,0,0); # where to go. 0: a gap at series2 (e.g. 1,1 -> 2,1); 1: a gap at series1 (e.g. 1,1 -> 1,2); 2: match (e.g. 1,1->2,2)
				$path_count[$i][$j]=\@counts;
			}
		}
		
		### alignment (time warping) with every marker gene ###
		for(my $idx=0; $idx<@{$series1[0]}; $idx++){
			next if($weights[$idx]==0); # if the weight of this marker is 0, skip it.
			my $alignment=time_warping(\@series1, \@series2, $w_size, $idx);
			my @aligned=@{$alignment};

			## transform alignment results (two vectors with the same length) into path ##
			for(my $i=0; $i<@{$aligned[0]}-1; $i++){
				my @this=($aligned[0][$i], $aligned[1][$i]);
				my @next=($aligned[0][$i+1], $aligned[1][$i+1]);
				my $direction=($next[0]==$this[0]+1 && $next[1]==$this[1]+1) ? 2 : ($this[0]==$next[0] ? 1 : 0 );
				
				$path_count[$this[0]][$this[1]][$direction]+=$weights[$idx];
			}
		}

		### start integrate ###
		my @align_final=();
		my ($i, $j)=(1, 1);
		my ($gap_num, $score)=(0,1);
		while($i<=@series1 && $j<=@series2){
			push(@{$align_final[0]}, $i);
			push(@{$align_final[1]}, $j);
			last if($i==@series1 && $j==@series2);

			my $direction=which_max(@{$path_count[$i][$j]});
			$score*=max(@{$path_count[$i][$j]}) / sum(@{$path_count[$i][$j]});
			if($direction==0){
				$gap_num++;
				$i++;
			}
			elsif($direction==1){
				$gap_num++;
				$j++;
			}
			else{
				$i++; $j++;
			}
		}
		return(\@align_final, $gap_num, $score);
	}
	# mode 1 (default): do alignment based on all markers together, by considering all of them when calculating distance
	elsif($align_mode==1){
		my $alignment=time_warping(\@series1, \@series2, $w_size, -1, \@weights);
		my @aligned=@{$alignment};
		my $gap_num=(scalar @{$aligned[0]} - max(@{$aligned[0]})) + (scalar @{$aligned[1]} - max(@{$aligned[1]}));
		
		return(\@aligned, $gap_num, -1);
	}
}



############################ Here start the main part of the script ################################
### Usage: ./align.pl [FILE:marker_profile] [FILE:weight] [alignment_mode] [assigned_template_index] [section_numbers]
##### alignment_mode: interger, either 0 or 1
##### assigned_template_index: interger, -1 for no assignment, 0-? for template index
##### section_numbers: the number of sections for each individual, joined by comma

my $file=shift @ARGV; # the expression profile of all markers
my $file_markers_weight=shift @ARGV; # the file with weight of each marker

my $mode=shift @ARGV; # alignment mode

my $assign_template=shift @ARGV; # the index of template individual; -1 for no pre-assigned reference
my @section_num=split(",", $ARGV[0]); # the number of sections in each individual

my $total_section_num=sum(@section_num);
my $individual_num=scalar @section_num;

warn "use assigned template: individual $assign_template\n" if($assign_template!=-1);


###### part 1. read marker weight file #########
my @weights=();
open(FILE, $file_markers_weight)||die($!);
while(<FILE>){
	chomp $_;
	push (@weights, $_);
}
close FILE;


###### part 2. read marker profiles #########
open(FILE, $file)||die($!);
my @values=(); # $values[num of samples][num of sections][num of markers]
my @sample_names=();
my $marker_count=-1;
while(<FILE>){
	chomp $_;
	$marker_count++;
	if($marker_count==0){
		@sample_names=split("\t");
		if(scalar @sample_names != $total_section_num){
			warn "Information given wrong. There should be ",$total_section_num," columns, instead of ",scalar @sample_names,", in one row.\n";
			exit;
		}
	}
	else{
		my ($marker_name, @elem)=split("\t");
		if(scalar @elem!=scalar @sample_names){
			warn "Sample number inconsistent. There should be ",$total_section_num," samples, but only ",scalar @elem," are available at line ",$marker_count+2,".\n";
			exit;
		}

		my $sample_count=0;
		my $section_count=0;
		foreach (@elem){
			$values[$sample_count][$section_count][$marker_count-1]=$_;
			$section_count++;
			if($section_count==$section_num[$sample_count]){
				$section_count=0;
				$sample_count++;
			}
		}
	}
}
close FILE;

if($marker_count!=scalar @weights){
	warn "Error! The number of markers does not fit the number of weights.\n";
	exit;
}

##### part 3. align #########
my @correspond_markers=();
my @alignment_results=();
my $fewest_gap_num=-1; my $best_template=-1; my @best_scores=(); my $best_score=0;

warn "start alignment.\n";
if($assign_template == -1){
	for(my $template=0; $template<$individual_num; $template++){
		my $count_gap=0;
		my @scores=();
		for(my $i=0; $i<$individual_num; $i++){
			next if($i==$template);
			
			my @section_num_sorted=sort{$a<=>$b} @section_num;
			my $max_section_num_diff=$section_num_sorted[-1]-$section_num_sorted[0];
			my @align_res=align($values[$template], $values[$i], \@weights, $mode, max(5,$max_section_num_diff));
			
			@{$alignment_results[$template][$i]}=@{$align_res[0]};
			$count_gap+=$align_res[1];
			push(@scores, $align_res[2]);
		}
		my $score=sum(@scores);
	
		#if($fewest_gap_num==-1 || $fewest_gap_num>$count_gap || ($fewest_gap_num==$count_gap && $section_num[$template]>$section_num[$best_template])  ||($fewest_gap_num==$count_gap && $section_num[$template]==$section_num[$best_template] && $best_score<$score)){
		if($fewest_gap_num==-1 || $section_num[$template]>$section_num[$best_template] || ($section_num[$template]==$section_num[$best_template] && $fewest_gap_num>$count_gap) || ($fewest_gap_num==$count_gap && $section_num[$template]==$section_num[$best_template] && $best_score<$score)){
			$fewest_gap_num=$count_gap;
			$best_template=$template;
			@best_scores=@scores;
			$best_score=$score;
		}
		warn "$template\t",$count_gap,"\t",join(",", @scores),"\n";
	}
}
else{
	my $template=$assign_template;
	my $count_gap=0;
	my @scores=();
	for(my $i=0; $i<$individual_num; $i++){
		next if($i==$template);
		
		my @align_res=align($values[$template], $values[$i], \@weights, $mode, 5);
		@{$alignment_results[$template][$i]}=@{$align_res[0]};
		$count_gap+=$align_res[1];
		push(@scores, $align_res[2]);
	}
	
	$best_template=$template; $fewest_gap_num=$count_gap; @best_scores=@scores;
	warn "$template\t",$count_gap,"\t",join(",", @scores),"\n";
}


######## part 4. output ############
print "#template: $best_template; $fewest_gap_num; ",join(",", @best_scores),"\n";
for(my $i=0; $i<$individual_num; $i++){
	next if($i==$best_template);
	print ">$best_template:$i\n";
	print join("\t", @{$alignment_results[$best_template][$i][0]}),"\n";
	print join("\t", @{$alignment_results[$best_template][$i][1]}),"\n";
}

exit;
